@@ -9,15 +9,17 @@
 #include "frameobject.h"
 #include "expat.h"
 
+#define XML_COMBINED_VERSION (10000*XML_MAJOR_VERSION+100*XML_MINOR_VERSION+XML_MICRO_VERSION)
+
 #ifndef PyDoc_STRVAR
 
 /*
  * fdrake says:
  * Don't change the PyDoc_STR macro definition to (str), because
- * '''the parentheses cause compile failures 
- * ("non-constant static initializer" or something like that) 
+ * '''the parentheses cause compile failures
+ * ("non-constant static initializer" or something like that)
  * on some platforms (Irix?)'''
- */ 
+ */
 #define PyDoc_STR(str)         str
 #define PyDoc_VAR(name)        static char name[]
 #define PyDoc_STRVAR(name,str) PyDoc_VAR(name) = PyDoc_STR(str)
@@ -51,7 +53,9 @@ enum HandlerTypes {
     XmlDecl,
     ElementDecl,
     AttlistDecl,
+#if XML_COMBINED_VERSION >= 19504
     SkippedEntity,
+#endif
     _DummyDecl
 };
 
@@ -81,7 +85,7 @@ typedef struct {
 
 #define CHARACTER_DATA_BUFFER_SIZE 8192
 
-staticforward PyTypeObject Xmlparsetype;
+static PyTypeObject Xmlparsetype;
 
 typedef void (*xmlhandlersetter)(XML_Parser self, void *meth);
 typedef void* xmlhandler;
@@ -94,7 +98,7 @@ struct HandlerInfo {
     PyObject *nameobj;
 };
 
-staticforward struct HandlerInfo handler_info[64];
+static struct HandlerInfo handler_info[64];
 
 /* Set an integer attribute on the error object; return true on success,
  * false on an exception.
@@ -221,10 +225,25 @@ conv_string_len_to_utf8(const XML_Char *str, int len)
 
 static void clear_handlers(xmlparseobject *self, int initial);
 
+/* This handler is used when an error has been detected, in the hope
+   that actual parsing can be terminated early.  This will only help
+   if an external entity reference is encountered. */
+static int
+error_external_entity_ref_handler(XML_Parser parser,
+                                  const XML_Char *context,
+                                  const XML_Char *base,
+                                  const XML_Char *systemId,
+                                  const XML_Char *publicId)
+{
+    return 0;
+}
+
 static void
 flag_error(xmlparseobject *self)
 {
     clear_handlers(self, 0);
+    XML_SetExternalEntityRefHandler(self->itself,
+                                    error_external_entity_ref_handler);
 }
 
 static PyCodeObject*
@@ -320,7 +339,7 @@ call_with_frame(PyCodeObject *c, PyObject* func, PyObject* args)
     f = PyFrame_New(
                     tstate,			/*back*/
                     c,				/*code*/
-                    tstate->frame->f_globals,	/*globals*/
+                    PyEval_GetGlobals(),	/*globals*/
                     NULL			/*locals*/
                     );
     if (f == NULL)
@@ -710,12 +729,14 @@ VOID_HANDLER(AttlistDecl,
               STRING_CONV_FUNC,att_type, STRING_CONV_FUNC,dflt,
               isrequired))
 
+#if XML_COMBINED_VERSION >= 19504
 VOID_HANDLER(SkippedEntity,
              (void *userData,
               const XML_Char *entityName,
               int is_parameter_entity),
              ("Ni",
               string_intern(self, entityName), is_parameter_entity))
+#endif
 
 VOID_HANDLER(NotationDecl,
 		(void *userData,
@@ -856,7 +877,12 @@ readinst(char *buf, int buf_size, PyObject *meth)
 
     PyTuple_SET_ITEM(arg, 0, bytes);
 
-    if ((str = PyObject_CallObject(meth, arg)) == NULL)
+#if PY_VERSION_HEX < 0x02020000
+    str = PyObject_CallObject(meth, arg);
+#else
+    str = PyObject_Call(meth, arg, NULL);
+#endif
+    if (str == NULL)
         goto finally;
 
     /* XXX what to do if it returns a Unicode string? */
@@ -1111,6 +1137,8 @@ xmlparse_SetParamEntityParsing(xmlparseobject *p, PyObject* args)
     return PyInt_FromLong(flag);
 }
 
+
+#if XML_COMBINED_VERSION >= 19505
 PyDoc_STRVAR(xmlparse_UseForeignDTD__doc__,
 "UseForeignDTD([flag])\n\
 Allows the application to provide an artificial external subset if one is\n\
@@ -1136,6 +1164,7 @@ xmlparse_UseForeignDTD(xmlparseobject *self, PyObject *args)
     Py_INCREF(Py_None);
     return Py_None;
 }
+#endif
 
 static struct PyMethodDef xmlparse_methods[] = {
     {"Parse",	  (PyCFunction)xmlparse_Parse,
@@ -1152,8 +1181,10 @@ static struct PyMethodDef xmlparse_methods[] = {
 		  METH_VARARGS, xmlparse_SetParamEntityParsing__doc__},
     {"GetInputContext", (PyCFunction)xmlparse_GetInputContext,
 		  METH_VARARGS, xmlparse_GetInputContext__doc__},
+#if XML_COMBINED_VERSION >= 19505
     {"UseForeignDTD", (PyCFunction)xmlparse_UseForeignDTD,
 		  METH_VARARGS, xmlparse_UseForeignDTD__doc__},
+#endif
     {NULL,	  NULL}		/* sentinel */
 };
 
@@ -1391,25 +1422,37 @@ xmlparse_getattr(xmlparseobject *self, char *name)
         }
     }
 
+#define APPEND(list, str)				\
+        do {						\
+                PyObject *o = PyString_FromString(str);	\
+                if (o != NULL)				\
+        	        PyList_Append(list, o);		\
+                Py_XDECREF(o);				\
+        } while (0)
+
     if (strcmp(name, "__members__") == 0) {
         int i;
         PyObject *rc = PyList_New(0);
         for (i = 0; handler_info[i].name != NULL; i++) {
-            PyList_Append(rc, get_handler_name(&handler_info[i]));
+            PyObject *o = get_handler_name(&handler_info[i]);
+            if (o != NULL)
+                PyList_Append(rc, o);
+            Py_XDECREF(o);
         }
-        PyList_Append(rc, PyString_FromString("ErrorCode"));
-        PyList_Append(rc, PyString_FromString("ErrorLineNumber"));
-        PyList_Append(rc, PyString_FromString("ErrorColumnNumber"));
-        PyList_Append(rc, PyString_FromString("ErrorByteIndex"));
-        PyList_Append(rc, PyString_FromString("buffer_size"));
-        PyList_Append(rc, PyString_FromString("buffer_text"));
-        PyList_Append(rc, PyString_FromString("buffer_used"));
-        PyList_Append(rc, PyString_FromString("namespace_prefixes"));
-        PyList_Append(rc, PyString_FromString("ordered_attributes"));
-        PyList_Append(rc, PyString_FromString("returns_unicode"));
-        PyList_Append(rc, PyString_FromString("specified_attributes"));
-        PyList_Append(rc, PyString_FromString("intern"));
+        APPEND(rc, "ErrorCode");
+        APPEND(rc, "ErrorLineNumber");
+        APPEND(rc, "ErrorColumnNumber");
+        APPEND(rc, "ErrorByteIndex");
+        APPEND(rc, "buffer_size");
+        APPEND(rc, "buffer_text");
+        APPEND(rc, "buffer_used");
+        APPEND(rc, "namespace_prefixes");
+        APPEND(rc, "ordered_attributes");
+        APPEND(rc, "returns_unicode");
+        APPEND(rc, "specified_attributes");
+        APPEND(rc, "intern");
 
+#undef APPEND
         return rc;
     }
     return Py_FindMethod(xmlparse_methods, (PyObject *)self, name);
@@ -1667,7 +1710,7 @@ PyDoc_STRVAR(pyexpat_module_documentation,
 static PyObject *
 get_version_string(void)
 {
-    static char *rcsid = "#Revision: 2.64 $";
+    static char *rcsid = "#Revision: 2.78 $";
     char *rev = rcsid;
     int i = 0;
 
@@ -1689,9 +1732,17 @@ get_version_string(void)
 #define MODULE_INITFUNC initpyexpat
 #endif
 
-void MODULE_INITFUNC(void);  /* avoid compiler warnings */
+#ifndef PyMODINIT_FUNC
+#   ifdef MS_WINDOWS
+#       define PyMODINIT_FUNC __declspec(dllexport) void
+#   else
+#       define PyMODINIT_FUNC void
+#   endif
+#endif
 
-DL_EXPORT(void)
+PyMODINIT_FUNC MODULE_INITFUNC(void);  /* avoid compiler warnings */
+
+PyMODINIT_FUNC
 MODULE_INITFUNC(void)
 {
     PyObject *m, *d;
@@ -1746,7 +1797,7 @@ MODULE_INITFUNC(void)
     PyModule_AddStringConstant(m, "native_encoding", "UTF-8");
 
     /* THIS IS FOR USE IN PyXML ONLY.  */
-    PyModule_AddStringConstant(m, "pyxml_expat_version", "$Revision: 1.68 $");
+    PyModule_AddStringConstant(m, "pyxml_expat_version", "$Revision: 1.74 $");
 
     sys_modules = PySys_GetObject("modules");
     d = PyModule_GetDict(m);
@@ -1774,6 +1825,7 @@ MODULE_INITFUNC(void)
         /* Don't core dump later! */
         return;
     
+#if XML_COMBINED_VERSION > 19505
     {
         const XML_Feature *features = XML_GetFeatureList();
         PyObject *list = PyList_New(0);
@@ -1802,6 +1854,7 @@ MODULE_INITFUNC(void)
                 PyModule_AddObject(m, "features", list);
         }
     }
+#endif
 
 #define MYCONST(name) \
     PyModule_AddStringConstant(errors_module, #name, \
@@ -1877,7 +1930,7 @@ clear_handlers(xmlparseobject *self, int initial)
     }
 }
 
-statichere struct HandlerInfo handler_info[] = {
+static struct HandlerInfo handler_info[] = {
     {"StartElementHandler",
      (xmlhandlersetter)XML_SetStartElementHandler,
      (xmlhandler)my_StartElementHandler},
@@ -1941,9 +1994,11 @@ statichere struct HandlerInfo handler_info[] = {
     {"AttlistDeclHandler",
      (xmlhandlersetter)XML_SetAttlistDeclHandler,
      (xmlhandler)my_AttlistDeclHandler},
+#if XML_COMBINED_VERSION >= 19504
     {"SkippedEntityHandler",
      (xmlhandlersetter)XML_SetSkippedEntityHandler,
      (xmlhandler)my_SkippedEntityHandler},
+#endif
 
     {NULL, NULL, NULL} /* sentinel */
 };
