@@ -1,7 +1,8 @@
 # test for xml.dom.minidom
 
-import os.path
+import os
 import sys
+import pickle
 import traceback
 from StringIO import StringIO
 from test.test_support import verbose
@@ -13,15 +14,40 @@ import xml.parsers.expat
 from xml.dom.minidom import parse, Node, Document, parseString
 from xml.dom.minidom import getDOMImplementation
 
+try:
+    from os import extsep
+except ImportError:
+    extsep = '.'
 
 if __name__ == "__main__":
     base = sys.argv[0]
 else:
     base = __file__
-tstfile = os.path.join(os.path.dirname(base), "test.xml")
+tstfile = os.path.join(os.path.dirname(base), "test"+extsep+"xml")
 del base
 
-def confirm(test, testname="Test"):
+# Python 2.1 and earlier does not fully support the NodeList interface
+testlist = xml.dom.minidom.NodeList()
+try:
+    testlist.length
+except AttributeError:
+    def t_length(list, length):
+        return len(list) == length
+    def t_item(list, index, item):
+        if item is None:
+            # .item can go past the end, indexing cannot
+            return 1
+        return list[index] is item
+else:
+    def t_length(list, length):
+        return len(list) == length and list.length == length
+    def t_item(list, index, item):
+        if item is None:
+            return list.item(index) is item
+        return list[index] is item and list.item(index) is item
+del testlist
+
+def confirm(test, testname = "Test"):
     if not test:
         print "Failed " + testname
         raise Exception
@@ -43,40 +69,85 @@ def testInsertBefore():
     elem = root.childNodes[0]
     nelem = dom.createElement("element")
     root.insertBefore(nelem, elem)
-    confirm(len(root.childNodes) == 2
-            and root.childNodes[0] is nelem
-            and root.childNodes[1] is elem
+    confirm(t_length(root.childNodes, 2)
+            and t_item(root.childNodes, 0, nelem)
+            and t_item(root.childNodes, 1, elem)
             and root.firstChild is nelem
             and root.lastChild is elem
             and root.toxml() == "<doc><element/><foo/></doc>"
             , "testInsertBefore -- node properly placed in tree")
     nelem = dom.createElement("element")
     root.insertBefore(nelem, None)
-    confirm(len(root.childNodes) == 3
-            and root.childNodes[1] is elem
-            and root.childNodes[2] is nelem
+    confirm(t_length(root.childNodes, 3)
+            and t_item(root.childNodes, 1, elem)
+            and t_item(root.childNodes, 2, nelem)
             and root.lastChild is nelem
             and nelem.previousSibling is elem
             and root.toxml() == "<doc><element/><foo/><element/></doc>"
             , "testInsertBefore -- node properly placed in tree")
     nelem2 = dom.createElement("bar")
     root.insertBefore(nelem2, nelem)
-    confirm(len(root.childNodes) == 4
-            and root.childNodes[2] is nelem2
-            and root.childNodes[3] is nelem
+    confirm(t_length(root.childNodes, 4)
+            and t_item(root.childNodes, 2, nelem2)
+            and t_item(root.childNodes, 3, nelem)
             and nelem2.nextSibling is nelem
             and nelem.previousSibling is nelem2
             and root.toxml() == "<doc><element/><foo/><bar/><element/></doc>"
             , "testInsertBefore -- node properly placed in tree")
     dom.unlink()
 
+def _create_fragment_test_nodes():
+    dom = parseString("<doc/>")
+    orig = dom.createTextNode("original")
+    c1 = dom.createTextNode("foo")
+    c2 = dom.createTextNode("bar")
+    c3 = dom.createTextNode("bat")
+    dom.documentElement.appendChild(orig)
+    frag = dom.createDocumentFragment()
+    frag.appendChild(c1)
+    frag.appendChild(c2)
+    frag.appendChild(c3)
+    return dom, orig, c1, c2, c3, frag
+
+def testInsertBeforeFragment():
+    dom, orig, c1, c2, c3, frag = _create_fragment_test_nodes()
+    dom.documentElement.insertBefore(frag, None)
+    confirm(tuple(dom.documentElement.childNodes) == (orig, c1, c2, c3),
+            "insertBefore(<fragment>, None)")
+    frag.unlink()
+    dom.unlink()
+    #
+    dom, orig, c1, c2, c3, frag = _create_fragment_test_nodes()
+    dom.documentElement.insertBefore(frag, orig)
+    confirm(tuple(dom.documentElement.childNodes) == (c1, c2, c3, orig),
+            "insertBefore(<fragment>, orig)")
+    frag.unlink()
+    dom.unlink()
+
 def testAppendChild():
     dom = parse(tstfile)
-    dom.documentElement.appendChild(dom.createComment( "Hello"))
+    dom.documentElement.appendChild(dom.createComment(u"Hello"))
     confirm(dom.documentElement.childNodes[-1].nodeName == "#comment")
     confirm(dom.documentElement.childNodes[-1].data == "Hello")
     dom.unlink()
 
+def testAppendChildFragment():
+    dom, orig, c1, c2, c3, frag = _create_fragment_test_nodes()
+    dom.documentElement.appendChild(frag)
+    confirm(tuple(dom.documentElement.childNodes) == (orig, c1, c2, c3),
+            "appendChild(<fragment>)")
+    frag.unlink()
+    dom.unlink()
+
+def testReplaceChildFragment():
+    dom, orig, c1, c2, c3, frag = _create_fragment_test_nodes()
+    dom.documentElement.replaceChild(frag, orig)
+    orig.unlink()
+    confirm(tuple(dom.documentElement.childNodes) == (c1, c2, c3),
+            "replaceChild(<fragment>)")
+    frag.unlink()
+    dom.unlink()
+
 def testLegalChildren():
     dom = Document()
     elem = dom.createElement('element')
@@ -112,6 +183,23 @@ def testLegalChildren():
     elem.appendChild(text)
     dom.unlink()
 
+def testNamedNodeMapSetItem():
+    dom = Document()
+    elem = dom.createElement('element')
+    attrs = elem.attributes
+    attrs["foo"] = "bar"
+    a = attrs.item(0)
+    confirm(a.ownerDocument is dom,
+            "NamedNodeMap.__setitem__() sets ownerDocument")
+    confirm(a.ownerElement is elem,
+            "NamedNodeMap.__setitem__() sets ownerElement")
+    confirm(a.value == "bar",
+            "NamedNodeMap.__setitem__() sets value")
+    confirm(a.nodeValue == "bar",
+            "NamedNodeMap.__setitem__() sets nodeValue")
+    elem.unlink()
+    dom.unlink()
+
 def testNonZero():
     dom = parse(tstfile)
     confirm(dom)# should not be zero
@@ -134,6 +222,11 @@ def testAAA():
     el = dom.documentElement
     el.setAttribute("spam", "jam2")
     confirm(el.toxml() == '<abc spam="jam2"/>', "testAAA")
+    a = el.getAttributeNode("spam")
+    confirm(a.ownerDocument is dom,
+            "setAttribute() sets ownerDocument")
+    confirm(a.ownerElement is dom.documentElement,
+            "setAttribute() sets ownerElement")
     dom.unlink()
 
 def testAAB():
@@ -207,7 +300,8 @@ def testRemoveAttributeNode():
     confirm(len(child.attributes) == 1)
     node = child.getAttributeNode("spam")
     child.removeAttributeNode(node)
-    confirm(len(child.attributes) == 0)
+    confirm(len(child.attributes) == 0
+            and child.getAttributeNode("spam") is None)
 
     dom.unlink()
 
@@ -217,13 +311,36 @@ def testChangeAttr():
     el.setAttribute("spam", "jam")
     confirm(len(el.attributes) == 1)
     el.setAttribute("spam", "bam")
-    confirm(len(el.attributes) == 1)
+    # Set this attribute to be an ID and make sure that doesn't change
+    # when changing the value:
+    el.setIdAttribute("spam")
+    confirm(len(el.attributes) == 1
+            and el.attributes["spam"].value == "bam"
+            and el.attributes["spam"].nodeValue == "bam"
+            and el.getAttribute("spam") == "bam"
+            and el.getAttributeNode("spam").isId)
     el.attributes["spam"] = "ham"
-    confirm(len(el.attributes) == 1)
+    confirm(len(el.attributes) == 1
+            and el.attributes["spam"].value == "ham"
+            and el.attributes["spam"].nodeValue == "ham"
+            and el.getAttribute("spam") == "ham"
+            and el.attributes["spam"].isId)
     el.setAttribute("spam2", "bam")
-    confirm(len(el.attributes) == 2)
-    el.attributes[ "spam2"] = "bam2"
-    confirm(len(el.attributes) == 2)
+    confirm(len(el.attributes) == 2
+            and el.attributes["spam"].value == "ham"
+            and el.attributes["spam"].nodeValue == "ham"
+            and el.getAttribute("spam") == "ham"
+            and el.attributes["spam2"].value == "bam"
+            and el.attributes["spam2"].nodeValue == "bam"
+            and el.getAttribute("spam2") == "bam")
+    el.attributes["spam2"] = "bam2"
+    confirm(len(el.attributes) == 2
+            and el.attributes["spam"].value == "ham"
+            and el.attributes["spam"].nodeValue == "ham"
+            and el.getAttribute("spam") == "ham"
+            and el.attributes["spam2"].value == "bam2"
+            and el.attributes["spam2"].nodeValue == "bam2"
+            and el.getAttribute("spam2") == "bam2")
     dom.unlink()
 
 def testGetAttrList():
@@ -285,7 +402,7 @@ def testElementReprAndStr():
 # commented out until Fredrick's fix is checked in
 def _testElementReprAndStrUnicode():
     dom = Document()
-    el = dom.appendChild(dom.createElement( "abc"))
+    el = dom.appendChild(dom.createElement(u"abc"))
     string1 = repr(el)
     string2 = str(el)
     confirm(string1 == string2)
@@ -295,7 +412,7 @@ def _testElementReprAndStrUnicode():
 def _testElementReprAndStrUnicodeNS():
     dom = Document()
     el = dom.appendChild(
-        dom.createElementNS( "http://www.slashdot.org",  "slash:abc"))
+        dom.createElementNS(u"http://www.slashdot.org", u"slash:abc"))
     string1 = repr(el)
     string2 = str(el)
     confirm(string1 == string2)
@@ -304,7 +421,7 @@ def _testElementReprAndStrUnicodeNS():
 
 def testAttributeRepr():
     dom = Document()
-    el = dom.appendChild(dom.createElement( "abc"))
+    el = dom.appendChild(dom.createElement(u"abc"))
     node = el.setAttribute("abc", "def")
     confirm(str(node) == repr(node))
     dom.unlink()
@@ -318,7 +435,20 @@ def testWriteXML():
     dom.unlink()
     confirm(str == domstr)
 
-def testProcessingInstruction(): pass
+def testProcessingInstruction():
+    dom = parseString('<e><?mypi \t\n data \t\n ?></e>')
+    pi = dom.documentElement.firstChild
+    confirm(pi.target == "mypi"
+            and pi.data == "data \t\n "
+            and pi.nodeName == "mypi"
+            and pi.nodeType == Node.PROCESSING_INSTRUCTION_NODE
+            and pi.attributes is None
+            and not pi.hasChildNodes()
+            and len(pi.childNodes) == 0
+            and pi.firstChild is None
+            and pi.lastChild is None
+            and pi.localName is None
+            and pi.namespaceURI == xml.dom.EMPTY_NAMESPACE)
 
 def testProcessingInstructionRepr(): pass
 
@@ -361,6 +491,30 @@ def testAttrListKeys(): pass
 
 def testAttrListKeysNS(): pass
 
+def testRemoveNamedItem():
+    doc = parseString("<doc a=''/>")
+    e = doc.documentElement
+    attrs = e.attributes
+    a1 = e.getAttributeNode("a")
+    a2 = attrs.removeNamedItem("a")
+    confirm(a1.isSameNode(a2))
+    try:
+        attrs.removeNamedItem("a")
+    except xml.dom.NotFoundErr:
+        pass
+
+def testRemoveNamedItemNS():
+    doc = parseString("<doc xmlns:a='http://xml.python.org/' a:b=''/>")
+    e = doc.documentElement
+    attrs = e.attributes
+    a1 = e.getAttributeNodeNS("http://xml.python.org/", "b")
+    a2 = attrs.removeNamedItemNS("http://xml.python.org/", "b")
+    confirm(a1.isSameNode(a2))
+    try:
+        attrs.removeNamedItemNS("http://xml.python.org/", "b")
+    except xml.dom.NotFoundErr:
+        pass
+
 def testAttrListValues(): pass
 
 def testAttrListLength(): pass
@@ -389,7 +543,7 @@ def testHasChildNodes(): pass
 
 def testCloneElementShallow():
     dom, clone = _setupCloneElement(0)
-    confirm(len(clone.childNodes) == 0
+    confirm(t_length(clone.childNodes, 0)
             and clone.parentNode is None
             and clone.toxml() == '<doc attr="value"/>'
             , "testCloneElementShallow")
@@ -397,7 +551,7 @@ def testCloneElementShallow():
 
 def testCloneElementDeep():
     dom, clone = _setupCloneElement(1)
-    confirm(len(clone.childNodes) == 1
+    confirm(t_length(clone.childNodes, 1)
             and clone.parentNode is None
             and clone.toxml() == '<doc attr="value"><foo/></doc>'
             , "testCloneElementDeep")
@@ -640,9 +794,9 @@ def testNormalize():
     root = doc.documentElement
     root.appendChild(doc.createTextNode("first"))
     root.appendChild(doc.createTextNode("second"))
-    confirm(len(root.childNodes) == 2, "testNormalize -- preparation")
+    confirm(t_length(root.childNodes, 2), "testNormalize -- preparation")
     doc.normalize()
-    confirm(len(root.childNodes) == 1
+    confirm(t_length(root.childNodes, 1)
             and root.firstChild is root.lastChild
             and root.firstChild.data == "firstsecond"
             , "testNormalize -- result")
@@ -652,7 +806,7 @@ def testNormalize():
     root = doc.documentElement
     root.appendChild(doc.createTextNode(""))
     doc.normalize()
-    confirm(len(root.childNodes) == 0,
+    confirm(t_length(root.childNodes, 0),
             "testNormalize -- single empty node removed")
     doc.unlink()
 
@@ -685,6 +839,18 @@ def testParents():
 
     doc.unlink()
 
+def testNodeListItem():
+    doc = parseString("<doc><e/><e/></doc>")
+    children = doc.childNodes
+    docelem = children[0]
+    confirm(t_item(children, 0, children[0])
+            and t_item(children, 1, None)
+            and t_item(docelem.childNodes, 0, docelem.childNodes[0])
+            and t_item(docelem.childNodes, 1, docelem.childNodes[1])
+            and t_item(docelem.childNodes[0].childNodes, 0, None),
+            "test NodeList.item()")
+    doc.unlink()
+
 def testSAX2DOM():
     from xml.dom import pulldom
 
@@ -721,6 +887,14 @@ def testSAX2DOM():
 
     doc.unlink()
 
+def testEncodings():
+    doc = parseString('<foo>&#x20ac;</foo>')
+    confirm(doc.toxml() == u'<?xml version="1.0" ?>\n<foo>\u20ac</foo>'
+            and doc.toxml('utf-8') == '<?xml version="1.0" encoding="utf-8"?>\n<foo>\xe2\x82\xac</foo>'
+            and doc.toxml('iso-8859-15') == '<?xml version="1.0" encoding="iso-8859-15"?>\n<foo>\xa4</foo>',
+            "testEncodings - encoding EURO SIGN")
+    doc.unlink()
+
 class UserDataHandler:
     called = 0
     def handle(self, operation, key, data, src, dst):
@@ -977,6 +1151,187 @@ def testReplaceWholeText():
     confirm(text is None
             and len(elem.childNodes) == 2)
 
+def testSchemaType():
+    doc = parseString(
+        "<!DOCTYPE doc [\n"
+        "  <!ENTITY e1 SYSTEM 'http://xml.python.org/e1'>\n"
+        "  <!ENTITY e2 SYSTEM 'http://xml.python.org/e2'>\n"
+        "  <!ATTLIST doc id   ID       #IMPLIED \n"
+        "                ref  IDREF    #IMPLIED \n"
+        "                refs IDREFS   #IMPLIED \n"
+        "                enum (a|b)    #IMPLIED \n"
+        "                ent  ENTITY   #IMPLIED \n"
+        "                ents ENTITIES #IMPLIED \n"
+        "                nm   NMTOKEN  #IMPLIED \n"
+        "                nms  NMTOKENS #IMPLIED \n"
+        "                text CDATA    #IMPLIED \n"
+        "    >\n"
+        "]><doc id='name' notid='name' text='splat!' enum='b'"
+        "       ref='name' refs='name name' ent='e1' ents='e1 e2'"
+        "       nm='123' nms='123 abc' />")
+    elem = doc.documentElement
+    # We don't want to rely on any specific loader at this point, so
+    # just make sure we can get to all the names, and that the
+    # DTD-based namespace is right.  The names can vary by loader
+    # since each supports a different level of DTD information.
+    t = elem.schemaType
+    confirm(t.name is None
+            and t.namespace == xml.dom.EMPTY_NAMESPACE)
+    names = "id notid text enum ref refs ent ents nm nms".split()
+    for name in names:
+        a = elem.getAttributeNode(name)
+        t = a.schemaType
+        confirm(hasattr(t, "name")
+                and t.namespace == xml.dom.EMPTY_NAMESPACE)
+
+def testSetIdAttribute():
+    doc = parseString("<doc a1='v' a2='w'/>")
+    e = doc.documentElement
+    a1 = e.getAttributeNode("a1")
+    a2 = e.getAttributeNode("a2")
+    confirm(doc.getElementById("v") is None
+            and not a1.isId
+            and not a2.isId)
+    e.setIdAttribute("a1")
+    confirm(e.isSameNode(doc.getElementById("v"))
+            and a1.isId
+            and not a2.isId)
+    e.setIdAttribute("a2")
+    confirm(e.isSameNode(doc.getElementById("v"))
+            and e.isSameNode(doc.getElementById("w"))
+            and a1.isId
+            and a2.isId)
+    # replace the a1 node; the new node should *not* be an ID
+    a3 = doc.createAttribute("a1")
+    a3.value = "v"
+    e.setAttributeNode(a3)
+    confirm(doc.getElementById("v") is None
+            and e.isSameNode(doc.getElementById("w"))
+            and not a1.isId
+            and a2.isId
+            and not a3.isId)
+    # renaming an attribute should not affect it's ID-ness:
+    doc.renameNode(a2, xml.dom.EMPTY_NAMESPACE, "an")
+    confirm(e.isSameNode(doc.getElementById("w"))
+            and a2.isId)
+
+def testSetIdAttributeNS():
+    NS1 = "http://xml.python.org/ns1"
+    NS2 = "http://xml.python.org/ns2"
+    doc = parseString("<doc"
+                      " xmlns:ns1='" + NS1 + "'"
+                      " xmlns:ns2='" + NS2 + "'"
+                      " ns1:a1='v' ns2:a2='w'/>")
+    e = doc.documentElement
+    a1 = e.getAttributeNodeNS(NS1, "a1")
+    a2 = e.getAttributeNodeNS(NS2, "a2")
+    confirm(doc.getElementById("v") is None
+            and not a1.isId
+            and not a2.isId)
+    e.setIdAttributeNS(NS1, "a1")
+    confirm(e.isSameNode(doc.getElementById("v"))
+            and a1.isId
+            and not a2.isId)
+    e.setIdAttributeNS(NS2, "a2")
+    confirm(e.isSameNode(doc.getElementById("v"))
+            and e.isSameNode(doc.getElementById("w"))
+            and a1.isId
+            and a2.isId)
+    # replace the a1 node; the new node should *not* be an ID
+    a3 = doc.createAttributeNS(NS1, "a1")
+    a3.value = "v"
+    e.setAttributeNode(a3)
+    confirm(e.isSameNode(doc.getElementById("w")))
+    confirm(not a1.isId)
+    confirm(a2.isId)
+    confirm(not a3.isId)
+    confirm(doc.getElementById("v") is None)
+    # renaming an attribute should not affect it's ID-ness:
+    doc.renameNode(a2, xml.dom.EMPTY_NAMESPACE, "an")
+    confirm(e.isSameNode(doc.getElementById("w"))
+            and a2.isId)
+
+def testSetIdAttributeNode():
+    NS1 = "http://xml.python.org/ns1"
+    NS2 = "http://xml.python.org/ns2"
+    doc = parseString("<doc"
+                      " xmlns:ns1='" + NS1 + "'"
+                      " xmlns:ns2='" + NS2 + "'"
+                      " ns1:a1='v' ns2:a2='w'/>")
+    e = doc.documentElement
+    a1 = e.getAttributeNodeNS(NS1, "a1")
+    a2 = e.getAttributeNodeNS(NS2, "a2")
+    confirm(doc.getElementById("v") is None
+            and not a1.isId
+            and not a2.isId)
+    e.setIdAttributeNode(a1)
+    confirm(e.isSameNode(doc.getElementById("v"))
+            and a1.isId
+            and not a2.isId)
+    e.setIdAttributeNode(a2)
+    confirm(e.isSameNode(doc.getElementById("v"))
+            and e.isSameNode(doc.getElementById("w"))
+            and a1.isId
+            and a2.isId)
+    # replace the a1 node; the new node should *not* be an ID
+    a3 = doc.createAttributeNS(NS1, "a1")
+    a3.value = "v"
+    e.setAttributeNode(a3)
+    confirm(e.isSameNode(doc.getElementById("w")))
+    confirm(not a1.isId)
+    confirm(a2.isId)
+    confirm(not a3.isId)
+    confirm(doc.getElementById("v") is None)
+    # renaming an attribute should not affect it's ID-ness:
+    doc.renameNode(a2, xml.dom.EMPTY_NAMESPACE, "an")
+    confirm(e.isSameNode(doc.getElementById("w"))
+            and a2.isId)
+
+def testPickledDocument():
+    doc = parseString("<?xml version='1.0' encoding='us-ascii'?>\n"
+                      "<!DOCTYPE doc PUBLIC 'http://xml.python.org/public'"
+                      " 'http://xml.python.org/system' [\n"
+                      "  <!ELEMENT e EMPTY>\n"
+                      "  <!ENTITY ent SYSTEM 'http://xml.python.org/entity'>\n"
+                      "]><doc attr='value'> text\n"
+                      "<?pi sample?> <!-- comment --> <e/> </doc>")
+    s = pickle.dumps(doc)
+    doc2 = pickle.loads(s)
+    stack = [(doc, doc2)]
+    while stack:
+        n1, n2 = stack.pop()
+        confirm(n1.nodeType == n2.nodeType
+                and len(n1.childNodes) == len(n2.childNodes)
+                and n1.nodeName == n2.nodeName
+                and not n1.isSameNode(n2)
+                and not n2.isSameNode(n1))
+        if n1.nodeType == Node.DOCUMENT_TYPE_NODE:
+            len(n1.entities)
+            len(n2.entities)
+            len(n1.notations)
+            len(n2.notations)
+            confirm(len(n1.entities) == len(n2.entities)
+                    and len(n1.notations) == len(n2.notations))
+            for i in range(len(n1.notations)):
+                no1 = n1.notations.item(i)
+                no2 = n1.notations.item(i)
+                confirm(no1.name == no2.name
+                        and no1.publicId == no2.publicId
+                        and no1.systemId == no2.systemId)
+                statck.append((no1, no2))
+            for i in range(len(n1.entities)):
+                e1 = n1.entities.item(i)
+                e2 = n2.entities.item(i)
+                confirm(e1.notationName == e2.notationName
+                        and e1.publicId == e2.publicId
+                        and e1.systemId == e2.systemId)
+                stack.append((e1, e2))
+        if n1.nodeType != Node.DOCUMENT_NODE:
+            confirm(n1.ownerDocument.isSameNode(doc)
+                    and n2.ownerDocument.isSameNode(doc2))
+        for i in range(len(n1.childNodes)):
+            stack.append((n1.childNodes[i], n2.childNodes[i]))
+
 
 # --- MAIN PROGRAM
 
@@ -985,24 +1340,40 @@ names.sort()
 
 failed = []
 
+try:
+    Node.allnodes
+except AttributeError:
+    # We don't actually have the minidom from the standard library,
+    # but are picking up the PyXML version from site-packages.
+    def check_allnodes():
+        pass
+else:
+    def check_allnodes():
+        confirm(len(Node.allnodes) == 0,
+                "assertion: len(Node.allnodes) == 0")
+        if len(Node.allnodes):
+            print "Garbage left over:"
+            if verbose:
+                print Node.allnodes.items()[0:10]
+            else:
+                # Don't print specific nodes if repeatable results
+                # are needed
+                print len(Node.allnodes)
+        Node.allnodes = {}
+
 for name in names:
-    if name[:4]=="test":
+    if name.startswith("test"):
         func = globals()[name]
         try:
             func()
+            check_allnodes()
         except:
             failed.append(name)
-            oldstdout = sys.stdout
-            sys.stdout = StringIO()
-            try:
-                print "Test Failed: " + name
-                sys.stdout.flush()
-                traceback.print_exc()
-                print sys.exc_info()[1]
-                oldstdout.write(sys.stdout.getvalue())
-            finally:
-                sys.stdout = oldstdout
-            raise
+            print "Test Failed: ", name
+            sys.stdout.flush()
+            traceback.print_exception(*sys.exc_info())
+            print `sys.exc_info()[1]`
+            Node.allnodes = {}
 
 if failed:
     print "\n\n\n**** Check for failures in these tests:"
