@@ -8,7 +8,7 @@
 Components for reading XML files from PyExpat (Python 1.6, 2.0 or from PyXML).
 WWW: http://4suite.com/4DOM         e-mail: support@4suite.com
 
-Copyright (c) 2000 Fourthought Inc, USA.   All Rights Reserved.
+Copyright (c) 2000-2001 Fourthought Inc, USA.   All Rights Reserved.
 See  http://4suite.com/COPYRIGHT  for license and copyright information
 """
 
@@ -48,6 +48,7 @@ class Reader(reader.Reader):
 
     def initParser(self):
         self.parser = expat.ParserCreate()
+        self.parser.buffer_text = 1
         self.parser.StartElementHandler = self.startElement
         self.parser.EndElementHandler = self.endElement
         self.parser.CharacterDataHandler = self.characters
@@ -125,16 +126,17 @@ class Reader(reader.Reader):
             if local == 'xmlns':
                 if self._namespaces.has_key(prefix):
                     old_nss[prefix] = self._namespaces[prefix]
-                else:
-                    del_nss.append(prefix)
-                if (prefix or value):
+                    if value:
+                        self._namespaces[prefix] = attribs[curr_attrib_key]
+                    else:
+                        del self._namespaces[prefix]
+                elif value:
                     self._namespaces[prefix] = attribs[curr_attrib_key]
-                else:
-                    del self._namespaces[prefix]
+                    del_nss.append(prefix)
 
         self._namespaceStack.append((old_nss, del_nss))
         (prefix, local) = SplitQName(name)
-        nameSpace = self._namespaces.get(prefix, '')
+        nameSpace = self._namespaces.get(prefix, None)
 
         if self._ownerDoc:
             new_element = self._ownerDoc.createElementNS(
@@ -147,14 +149,17 @@ class Reader(reader.Reader):
 
         for curr_attrib_key,curr_attrib_value in attribs.items():
             (prefix, local) = SplitQName(curr_attrib_key)
+            qname = local
             if local == 'xmlns':
                 namespace = XMLNS_NAMESPACE
-                attr = self._ownerDoc.createAttributeNS(namespace,
-                                                         local + ':' + prefix)
+                if prefix:
+                    qname = local + ':' + prefix
+                attr = self._ownerDoc.createAttributeNS(namespace, qname)
             else:
-                namespace = prefix and self._namespaces.get(prefix, '') or ''
-                attr = self._ownerDoc.createAttributeNS(namespace,
-                                                         (prefix and prefix + ':' + local) or local)
+                namespace = prefix and self._namespaces.get(prefix, None) or None
+                if prefix:
+                    qname = prefix + ':' + local
+                attr = self._ownerDoc.createAttributeNS(namespace, qname)
             attr.value = curr_attrib_value
             new_element.setAttributeNodeNS(attr)
         self._nodeStack.append(new_element)
