@@ -16,15 +16,22 @@ Todo:
 
 import xml.dom
 
-from xml.dom import EMPTY_NAMESPACE, EMPTY_PREFIX
+from xml.dom import EMPTY_NAMESPACE, EMPTY_PREFIX, XMLNS_NAMESPACE, domreg
 from xml.dom.minicompat import *
 from xml.dom.xmlbuilder import DOMImplementationLS, DocumentLS
 
 _TupleType = type(())
 
+# This is used by the ID-cache invalidation checks; the list isn't
+# actually complete, since the nodes being checked will never be the
+# DOCUMENT_NODE or DOCUMENT_FRAGMENT_NODE.  (The node being checked is
+# the node being added or removed, not the node being modified.)
+#
+_nodeTypes_with_children = (xml.dom.Node.ELEMENT_NODE,
+                            xml.dom.Node.ENTITY_REFERENCE_NODE)
+
 
 class Node(xml.dom.Node, GetattrMagic):
-    _make_parent_nodes = 1
     namespaceURI = None # this is non-null only for elements and attributes
     parentNode = None
     ownerDocument = None
@@ -89,6 +96,8 @@ class Node(xml.dom.Node, GetattrMagic):
                 index = self.childNodes.index(refChild)
             except ValueError:
                 raise xml.dom.NotFoundErr()
+            if newChild.nodeType in _nodeTypes_with_children:
+                _clear_id_cache(self)
             self.childNodes.insert(index, newChild)
             newChild.nextSibling = refChild
             refChild.previousSibling = newChild
@@ -98,8 +107,7 @@ class Node(xml.dom.Node, GetattrMagic):
                 newChild.previousSibling = node
             else:
                 newChild.previousSibling = None
-            if self._make_parent_nodes:
-                newChild.parentNode = self
+            newChild.parentNode = self
         return newChild
 
     def appendChild(self, node):
@@ -111,6 +119,8 @@ class Node(xml.dom.Node, GetattrMagic):
         if node.nodeType not in self._child_node_types:
             raise xml.dom.HierarchyRequestErr(
                 "%s cannot be child of %s" % (repr(node), repr(self)))
+        elif node.nodeType in _nodeTypes_with_children:
+            _clear_id_cache(self)
         if node.parentNode is not None:
             node.parentNode.removeChild(node)
         _append_child(self, node)
@@ -134,9 +144,11 @@ class Node(xml.dom.Node, GetattrMagic):
         except ValueError:
             raise xml.dom.NotFoundErr()
         self.childNodes[index] = newChild
-        if self._make_parent_nodes:
-            newChild.parentNode = self
-            oldChild.parentNode = None
+        newChild.parentNode = self
+        oldChild.parentNode = None
+        if (newChild.nodeType in _nodeTypes_with_children
+            or oldChild.nodeType in _nodeTypes_with_children):
+            _clear_id_cache(self)
         newChild.nextSibling = oldChild.nextSibling
         newChild.previousSibling = oldChild.previousSibling
         oldChild.nextSibling = None
@@ -157,9 +169,10 @@ class Node(xml.dom.Node, GetattrMagic):
         if oldChild.previousSibling is not None:
             oldChild.previousSibling.nextSibling = oldChild.nextSibling
         oldChild.nextSibling = oldChild.previousSibling = None
+        if oldChild.nodeType in _nodeTypes_with_children:
+            _clear_id_cache(self)
 
-        if self._make_parent_nodes:
-            oldChild.parentNode = None
+        oldChild.parentNode = None
         return oldChild
 
     def normalize(self):
@@ -170,7 +183,7 @@ class Node(xml.dom.Node, GetattrMagic):
                 if data and L and L[-1].nodeType == child.nodeType:
                     # collapse text node
                     node = L[-1]
-                    node.data = node.nodeValue = node.data + child.data
+                    node.data = node.data + child.data
                     node.nextSibling = child.nextSibling
                     child.unlink()
                 elif data:
@@ -200,6 +213,10 @@ class Node(xml.dom.Node, GetattrMagic):
     def isSupported(self, feature, version):
         return self.ownerDocument.implementation.hasFeature(feature, version)
 
+    def _get_localName(self):
+        # Overridden in Element and Attr where localName can be Non-Null
+        return None
+
     # Node interfaces from Level 3 (WD 9 April 2002)
 
     def isSameNode(self, other):
@@ -211,10 +228,6 @@ class Node(xml.dom.Node, GetattrMagic):
         else:
             return None
 
-    def _get_localName(self):
-        # Overridden in Element and Attr where localName can be Non-Null
-        return None
-
     # The "user data" functions use a dictionary that is only present
     # if some user data has been set, so be careful not to assume it
     # exists.
@@ -262,7 +275,7 @@ class Node(xml.dom.Node, GetattrMagic):
 
 defproperty(Node, "firstChild", doc="First child node, or None.")
 defproperty(Node, "lastChild",  doc="Last child node, or None.")
-defproperty(Node, "localName", doc="Namespace-local name of this node.")
+defproperty(Node, "localName",  doc="Namespace-local name of this node.")
 
 
 def _append_child(self, node):
@@ -270,11 +283,18 @@ def _append_child(self, node):
     childNodes = self.childNodes
     if childNodes:
         last = childNodes[-1]
-        node.previousSibling = last
-        last.nextSibling = node
+        node.__dict__["previousSibling"] = last
+        last.__dict__["nextSibling"] = node
     childNodes.append(node)
-    if self._make_parent_nodes:
-        node.parentNode = self
+    node.__dict__["parentNode"] = self
+
+def _in_document(node):
+    # return True iff node is part of a document tree
+    while node is not None:
+        if node.nodeType == Node.DOCUMENT_NODE:
+            return True
+        node = node.parentNode
+    return False
 
 def _write_data(writer, data):
     "Writes datachars to writer."
@@ -322,6 +342,7 @@ class Attr(Node):
     attributes = None
     ownerElement = None
     specified = False
+    _is_id = False
 
     _child_node_types = (Node.TEXT_NODE, Node.ENTITY_REFERENCE_NODE)
 
@@ -352,16 +373,21 @@ class Attr(Node):
         d = self.__dict__
         if name in ("value", "nodeValue"):
             d["value"] = d["nodeValue"] = value
-            d["childNodes"][0].data = value
+            d2 = self.childNodes[0].__dict__
+            d2["data"] = d2["nodeValue"] = value
+            if self.ownerElement is not None:
+                _clear_id_cache(self.ownerElement)
         elif name in ("name", "nodeName"):
             d["name"] = d["nodeName"] = value
+            if self.ownerElement is not None:
+                _clear_id_cache(self.ownerElement)
         else:
             d[name] = value
 
-    def _set_prefix(self, value):
+    def _set_prefix(self, prefix):
         nsuri = self.namespaceURI
-        if value == "xmlns":
-            if self.namespaceURI and self.namespaceURI != XMLNS_NAMESPACE:
+        if prefix == "xmlns":
+            if nsuri and nsuri != XMLNS_NAMESPACE:
                 raise xml.dom.NamespaceErr(
                     "illegal use of 'xmlns' prefix for the wrong namespace")
         d = self.__dict__
@@ -369,15 +395,68 @@ class Attr(Node):
         if prefix is None:
             newName = self.localName
         else:
-            newName = "%s:%s" % (value, self.localName)
+            newName = "%s:%s" % (prefix, self.localName)
+        if self.ownerElement:
+            _clear_id_cache(self.ownerElement)
         d['nodeName'] = d['name'] = newName
 
     def _set_value(self, value):
         d = self.__dict__
         d['value'] = d['nodeValue'] = value
+        if self.ownerElement:
+            _clear_id_cache(self.ownerElement)
         self.childNodes[0].data = value
 
-defproperty(Attr, "localName", doc="Namespace-local name of this attribute.")
+    def unlink(self):
+        # This implementation does not call the base implementation
+        # since most of that is not needed, and the expense of the
+        # method call is not warranted.  We duplicate the removal of
+        # children, but that's all we needed from the base class.
+        elem = self.ownerElement
+        if elem is not None:
+            del elem._attrs[self.nodeName]
+            del elem._attrsNS[(self.namespaceURI, self.localName)]
+            if self._is_id:
+                self._is_id = False
+                elem._magic_id_nodes -= 1
+                self.ownerDocument._magic_id_count -= 1
+        for child in self.childNodes:
+            child.unlink()
+        del self.childNodes[:]
+
+    def _get_isId(self):
+        if self._is_id:
+            return True
+        doc = self.ownerDocument
+        elem = self.ownerElement
+        if doc is None or elem is None:
+            return False
+
+        info = doc._get_elem_info(elem)
+        if info is None:
+            return False
+        if self.namespaceURI:
+            return info.isIdNS(self.namespaceURI, self.localName)
+        else:
+            return info.isId(self.nodeName)
+
+    def _get_schemaType(self):
+        doc = self.ownerDocument
+        elem = self.ownerElement
+        if doc is None or elem is None:
+            return _no_type
+
+        info = doc._get_elem_info(elem)
+        if info is None:
+            return _no_type
+        if self.namespaceURI:
+            return info.getAttributeTypeNS(self.namespaceURI, self.localName)
+        else:
+            return info.getAttributeType(self.nodeName)
+
+defproperty(Attr, "isId",       doc="True if this attribute is an ID.")
+defproperty(Attr, "localName",  doc="Namespace-local name of this attribute.")
+defproperty(Attr, "schemaType", doc="Schema type for this attribute.")
 
 
 class NamedNodeMap(NewStyle, GetattrMagic):
@@ -432,7 +511,7 @@ class NamedNodeMap(NewStyle, GetattrMagic):
     def values(self):
         return self._attrs.values()
 
-    def get(self, name, value = None):
+    def get(self, name, value=None):
         return self._attrs.get(name, value)
 
     __len__ = _get_length
@@ -452,14 +531,18 @@ class NamedNodeMap(NewStyle, GetattrMagic):
     # same as set
     def __setitem__(self, attname, value):
         if isinstance(value, StringTypes):
-            node = Attr(attname)
+            try:
+                node = self._attrs[attname]
+            except KeyError:
+                node = Attr(attname)
+                node.ownerDocument = self._ownerElement.ownerDocument
+                self.setNamedItem(node)
             node.value = value
-            node.ownerDocument = self._ownerElement.ownerDocument
         else:
             if not isinstance(value, Attr):
                 raise TypeError, "value must be a string or Attr object"
             node = value
-        self.setNamedItem(node)
+            self.setNamedItem(node)
 
     def getNamedItem(self, name):
         try:
@@ -475,17 +558,27 @@ class NamedNodeMap(NewStyle, GetattrMagic):
 
     def removeNamedItem(self, name):
         n = self.getNamedItem(name)
-        del self._attrs[n.nodeName]
-        del self._attrsNS[(n.namespaceURI, n.localName)]
-        if n.__dict__.has_key('ownerElement'):
-            n.__dict__['ownerElement'] = None
+        if n is not None:
+            _clear_id_cache(self._ownerElement)
+            del self._attrs[n.nodeName]
+            del self._attrsNS[(n.namespaceURI, n.localName)]
+            if n.__dict__.has_key('ownerElement'):
+                n.__dict__['ownerElement'] = None
+            return n
+        else:
+            raise xml.dom.NotFoundErr()
 
     def removeNamedItemNS(self, namespaceURI, localName):
         n = self.getNamedItemNS(namespaceURI, localName)
-        del self._attrs[n.nodeName]
-        del self._attrsNS[(n.namespaceURI, n.localName)]
-        if n.__dict__.has_key('ownerElement'):
-            n.__dict__['ownerElement'] = None
+        if n is not None:
+            _clear_id_cache(self._ownerElement)
+            del self._attrsNS[(n.namespaceURI, n.localName)]
+            del self._attrs[n.nodeName]
+            if n.__dict__.has_key('ownerElement'):
+                n.__dict__['ownerElement'] = None
+            return n
+        else:
+            raise xml.dom.NotFoundErr()
 
     def setNamedItem(self, node):
         if not isinstance(node, Attr):
@@ -497,6 +590,7 @@ class NamedNodeMap(NewStyle, GetattrMagic):
         self._attrs[node.name] = node
         self._attrsNS[(node.namespaceURI, node.localName)] = node
         node.ownerElement = self._ownerElement
+        _clear_id_cache(node.ownerElement)
         return old
 
     def setNamedItemNS(self, node):
@@ -504,9 +598,14 @@ class NamedNodeMap(NewStyle, GetattrMagic):
 
     def __delitem__(self, attname_or_tuple):
         node = self[attname_or_tuple]
+        _clear_id_cache(node.ownerElement)
         node.unlink()
-        del self._attrs[node.name]
-        del self._attrsNS[(node.namespaceURI, node.localName)]
+
+    def __getstate__(self):
+        return self._attrs, self._attrsNS, self._ownerElement
+
+    def __setstate__(self, state):
+        self._attrs, self._attrsNS, self._ownerElement = state
 
 defproperty(NamedNodeMap, "length",
             doc="Number of nodes in the NamedNodeMap.")
@@ -514,9 +613,34 @@ defproperty(NamedNodeMap, "length",
 AttributeList = NamedNodeMap
 
 
+class TypeInfo(NewStyle):
+    __slots__ = 'namespace', 'name'
+
+    def __init__(self, namespace, name):
+        self.namespace = namespace
+        self.name = name
+
+    def __repr__(self):
+        if self.namespace:
+            return "<TypeInfo %s (from %s)>" % (`self.name`, `self.namespace`)
+        else:
+            return "<TypeInfo %s>" % `self.name`
+
+    def _get_name(self):
+        return self.name
+
+    def _get_namespace(self):
+        return self.namespace
+
+_no_type = TypeInfo(None, None)
+
 class Element(Node):
     nodeType = Node.ELEMENT_NODE
     nodeValue = None
+    schemaType = _no_type
+
+    _magic_id_nodes = 0
+
     _child_node_types = (Node.ELEMENT_NODE,
                          Node.PROCESSING_INSTRUCTION_NODE,
                          Node.COMMENT_NODE,
@@ -565,21 +689,41 @@ class Element(Node):
             return ""
 
     def setAttribute(self, attname, value):
-        attr = Attr(attname)
-        # for performance
-        d = attr.__dict__
-        d["value"] = d["nodeValue"] = value
-        d["ownerDocument"] = self.ownerDocument
-        self.setAttributeNode(attr)
+        attr = self.getAttributeNode(attname)
+        if attr is None:
+            attr = Attr(attname)
+            # for performance
+            d = attr.__dict__
+            d["value"] = d["nodeValue"] = value
+            d["ownerDocument"] = self.ownerDocument
+            self.setAttributeNode(attr)
+        elif value != attr.value:
+            d = attr.__dict__
+            d["value"] = d["nodeValue"] = value
+            if attr.isId:
+                _clear_id_cache(self)
 
     def setAttributeNS(self, namespaceURI, qualifiedName, value):
         prefix, localname = _nssplit(qualifiedName)
-        # for performance
-        attr = Attr(qualifiedName, namespaceURI, localname, prefix)
-        d = attr.__dict__
-        d["value"] = d["nodeValue"] = value
-        d["ownerDocument"] = self.ownerDocument
-        self.setAttributeNode(attr)
+        attr = self.getAttributeNodeNS(namespaceURI, localname)
+        if attr is None:
+            # for performance
+            attr = Attr(qualifiedName, namespaceURI, localname, prefix)
+            d = attr.__dict__
+            d["prefix"] = prefix
+            d["nodeName"] = qualifiedName
+            d["value"] = d["nodeValue"] = value
+            d["ownerDocument"] = self.ownerDocument
+            self.setAttributeNode(attr)
+        else:
+            d = attr.__dict__
+            if value != attr.value:
+                d["value"] = d["nodeValue"] = value
+                if attr.isId:
+                    _clear_id_cache(self)
+            if attr.prefix != prefix:
+                d["prefix"] = prefix
+                d["nodeName"] = qualifiedName
 
     def getAttributeNode(self, attrname):
         return self._attrs.get(attrname)
@@ -590,32 +734,45 @@ class Element(Node):
     def setAttributeNode(self, attr):
         if attr.ownerElement not in (None, self):
             raise xml.dom.InuseAttributeErr("attribute node already owned")
-        old = self._attrs.get(attr.name, None)
-        if old:
-            old.unlink()
+        old1 = self._attrs.get(attr.name, None)
+        if old1 is not None:
+            self.removeAttributeNode(old1)
+        old2 = self._attrsNS.get((attr.namespaceURI, attr.localName), None)
+        if old2 is not None and old2 is not old1:
+            self.removeAttributeNode(old2)
         _set_attribute_node(self, attr)
 
-        if old is not attr:
+        if old1 is not attr:
             # It might have already been part of this node, in which case
             # it doesn't represent a change, and should not be returned.
-            return old
+            return old1
+        if old2 is not attr:
+            return old2
 
     setAttributeNodeNS = setAttributeNode
 
     def removeAttribute(self, name):
-        attr = self._attrs[name]
+        try:
+            attr = self._attrs[name]
+        except KeyError:
+            raise xml.dom.NotFoundErr()
         self.removeAttributeNode(attr)
 
     def removeAttributeNS(self, namespaceURI, localName):
-        attr = self._attrsNS[(namespaceURI, localName)]
+        try:
+            attr = self._attrsNS[(namespaceURI, localName)]
+        except KeyError:
+            raise xml.dom.NotFoundErr()
         self.removeAttributeNode(attr)
 
     def removeAttributeNode(self, node):
+        if node is None:
+            raise xml.dom.NotFoundErr()
         try:
-            del self._attrs[node.name]
+            self._attrs[node.name]
         except KeyError:
             raise xml.dom.NotFoundErr()
-        del self._attrsNS[(node.namespaceURI, node.localName)]
+        _clear_id_cache(self)
         node.unlink()
         # Restore this since the node is still useful and otherwise
         # unlinked
@@ -637,7 +794,7 @@ class Element(Node):
             self, namespaceURI, localName, NodeList())
 
     def __repr__(self):
-        return "<DOM Element: %s at %s>" % (self.tagName, id(self))
+        return "<DOM Element: %s at %#x>" % (self.tagName, id(self))
 
     def writexml(self, writer, indent="", addindent="", newl=""):
         # indent = current indentation
@@ -670,20 +827,42 @@ class Element(Node):
         else:
             return False
 
+    # DOM Level 3 attributes, based on the 22 Oct 2002 draft
+
+    def setIdAttribute(self, name):
+        idAttr = self.getAttributeNode(name)
+        self.setIdAttributeNode(idAttr)
+
+    def setIdAttributeNS(self, namespaceURI, localName):
+        idAttr = self.getAttributeNodeNS(namespaceURI, localName)
+        self.setIdAttributeNode(idAttr)
+
+    def setIdAttributeNode(self, idAttr):
+        if idAttr is None or not self.isSameNode(idAttr.ownerElement):
+            raise xml.dom.NotFoundErr()
+        if _get_containing_entref(self) is not None:
+            raise xml.dom.NoModificationAllowedErr()
+        if not idAttr._is_id:
+            idAttr.__dict__['_is_id'] = True
+            self._magic_id_nodes += 1
+            self.ownerDocument._magic_id_count += 1
+            _clear_id_cache(self)
+
 defproperty(Element, "attributes",
             doc="NamedNodeMap of attributes on the element.")
 defproperty(Element, "localName",
             doc="Namespace-local name of this element.")
 
 
-def _set_attribute_node(self, attr):
-    self._attrs[attr.name] = attr
-    self._attrsNS[(attr.namespaceURI, attr.localName)] = attr
+def _set_attribute_node(element, attr):
+    _clear_id_cache(element)
+    element._attrs[attr.name] = attr
+    element._attrsNS[(attr.namespaceURI, attr.localName)] = attr
 
     # This creates a circular reference, but Element.unlink()
     # breaks the cycle since the references to the attribute
     # dictionaries are tossed.
-    attr.__dict__['ownerElement'] = self
+    attr.__dict__['ownerElement'] = element
 
 
 class Childless:
@@ -830,7 +1009,8 @@ defproperty(CharacterData, "length", doc="Length of the string data.")
 class Text(CharacterData):
     # Make sure we don't add an instance __dict__ if we don't already
     # have one, at least when that's possible:
-    __slots__ = ()
+    # XXX this does not work, CharacterData is an old-style class
+    # __slots__ = ()
 
     nodeType = Node.TEXT_NODE
     nodeName = "#text"
@@ -904,10 +1084,42 @@ class Text(CharacterData):
         else:
             return None
 
+    def _get_isWhitespaceInElementContent(self):
+        if self.data.strip():
+            return False
+        elem = _get_containing_element(self)
+        if elem is None:
+            return False
+        info = self.ownerDocument._get_elem_info(elem)
+        if info is None:
+            return False
+        else:
+            return info.isElementContent()
 
+defproperty(Text, "isWhitespaceInElementContent",
+            doc="True iff this text node contains only whitespace"
+                " and is in element content.")
 defproperty(Text, "wholeText",
             doc="The text of all logically-adjacent text nodes.")
 
+
+def _get_containing_element(node):
+    c = node.parentNode
+    while c is not None:
+        if c.nodeType == Node.ELEMENT_NODE:
+            return c
+        c = c.parentNode
+    return None
+
+def _get_containing_entref(node):
+    c = node.parentNode
+    while c is not None:
+        if c.nodeType == Node.ENTITY_REFERENCE_NODE:
+            return c
+        c = c.parentNode
+    return None
+
+
 class Comment(Childless, CharacterData):
     nodeType = Node.COMMENT_NODE
     nodeName = "#comment"
@@ -922,7 +1134,8 @@ class Comment(Childless, CharacterData):
 class CDATASection(Text):
     # Make sure we don't add an instance __dict__ if we don't already
     # have one, at least when that's possible:
-    __slots__ = ()
+    # XXX this does not work, Text is an old-style class
+    # __slots__ = ()
 
     nodeType = Node.CDATA_SECTION_NODE
     nodeName = "#cdata-section"
@@ -958,9 +1171,12 @@ class ReadOnlySequentialNamedNodeMap(NewStyle, GetattrMagic):
 
     def __getitem__(self, name_or_tuple):
         if isinstance(name_or_tuple, _TupleType):
-            return self.getNamedItemNS(*name_or_tuple)
+            node = self.getNamedItemNS(*name_or_tuple)
         else:
-            return self.getNamedItem(name_or_tuple)
+            node = self.getNamedItem(name_or_tuple)
+        if node is None:
+            raise KeyError, name_or_tuple
+        return node
 
     def item(self, index):
         if index < 0:
@@ -986,6 +1202,12 @@ class ReadOnlySequentialNamedNodeMap(NewStyle, GetattrMagic):
         raise xml.dom.NoModificationAllowedErr(
             "NamedNodeMap instance is read-only")
 
+    def __getstate__(self):
+        return [self._seq]
+
+    def __setstate__(self, state):
+        self._seq = state[0]
+
 defproperty(ReadOnlySequentialNamedNodeMap, "length",
             doc="Number of entries in the NamedNodeMap.")
 
@@ -993,7 +1215,8 @@ defproperty(ReadOnlySequentialNamedNodeMap, "length",
 class Identified:
     """Mix-in class that supports the publicId and systemId attributes."""
 
-    __slots__ = 'publicId', 'systemId'
+    # XXX this does not work, this is an old-style class
+    # __slots__ = 'publicId', 'systemId'
 
     def _identified_mixin_init(self, publicId, systemId):
         self.publicId = publicId
@@ -1089,6 +1312,22 @@ class Entity(Identified, Node):
     def _get_version(self):
         return self.version
 
+    def appendChild(self, newChild):
+        raise xml.dom.HierarchyRequestErr(
+            "cannot append children to an entity node")
+
+    def insertBefore(self, newChild, refChild):
+        raise xml.dom.HierarchyRequestErr(
+            "cannot insert children below an entity node")
+
+    def removeChild(self, oldChild):
+        raise xml.dom.HierarchyRequestErr(
+            "cannot remove children from an entity node")
+
+    def replaceChild(self, newChild, oldChild):
+        raise xml.dom.HierarchyRequestErr(
+            "cannot replace children of an entity node")
+
 class Notation(Identified, Childless, Node):
     nodeType = Node.NOTATION_NODE
     nodeValue = None
@@ -1193,6 +1432,15 @@ class ElementInfo(NewStyle):
     def __init__(self, name):
         self.tagName = name
 
+    def getAttributeType(self, aname):
+        return _no_type
+
+    def getAttributeTypeNS(self, namespaceURI, localName):
+        return _no_type
+
+    def isElementContent(self):
+        return False
+
     def isEmpty(self):
         """Returns true iff this element is declared to have an EMPTY
         content model."""
@@ -1206,6 +1454,20 @@ class ElementInfo(NewStyle):
         """Returns true iff the identified attribute is a DTD-style ID."""
         return False
 
+    def __getstate__(self):
+        return self.tagName
+
+    def __setstate__(self, state):
+        self.tagName = state
+
+def _clear_id_cache(node):
+    if node.nodeType == Node.DOCUMENT_NODE:
+        node._id_cache.clear()
+        node._id_search_stack = None
+    elif _in_document(node):
+        node.ownerDocument._id_cache.clear()
+        node.ownerDocument._id_search_stack= None
+
 class Document(Node, DocumentLS):
     _child_node_types = (Node.ELEMENT_NODE, Node.PROCESSING_INSTRUCTION_NODE,
                          Node.COMMENT_NODE, Node.DOCUMENT_TYPE_NODE)
@@ -1230,10 +1492,22 @@ class Document(Node, DocumentLS):
     errorHandler = None
     documentURI = None
 
+    _magic_id_count = 0
+
     def __init__(self):
         self.childNodes = NodeList()
-        # mapping of (namespaceURI, tagName) -> ElementInfo
+        # mapping of (namespaceURI, localName) -> ElementInfo
+        #        and tagName -> ElementInfo
         self._elem_info = {}
+        self._id_cache = {}
+        self._id_search_stack = None
+
+    def _get_elem_info(self, element):
+        if element.namespaceURI:
+            key = element.namespaceURI, element.localName
+        else:
+            key = element.tagName
+        return self._elem_info.get(key)
 
     def _get_actualEncoding(self):
         return self.actualEncoding
@@ -1264,6 +1538,9 @@ class Document(Node, DocumentLS):
             raise xml.dom.HierarchyRequestErr(
                 "%s cannot be child of %s" % (repr(node), repr(self)))
         if node.parentNode is not None:
+            # This needs to be done before the next test since this
+            # may *be* the document element, in which case it should
+            # end up re-ordered to the end.
             node.parentNode.removeChild(node)
 
         if node.nodeType == Node.ELEMENT_NODE \
@@ -1371,36 +1648,76 @@ class Document(Node, DocumentLS):
         a.value = ""
         return a
 
+    # A couple of implementation-specific helpers to create node types
+    # not supported by the W3C DOM specs:
+
+    def _create_entity(self, name, publicId, systemId, notationName):
+        e = Entity(name, publicId, systemId, notationName)
+        e.ownerDocument = self
+        return e
+
+    def _create_notation(self, name, publicId, systemId):
+        n = Notation(name, publicId, systemId)
+        n.ownerDocument = self
+        return n
+
     def getElementById(self, id):
-        if not self._elem_info:
+        if self._id_cache.has_key(id):
+            return self._id_cache[id]
+        if not (self._elem_info or self._magic_id_count):
             return None
-        stack = [self.documentElement]
+
+        stack = self._id_search_stack
+        if stack is None:
+            # we never searched before, or the cache has been cleared
+            stack = [self.documentElement]
+            self._id_search_stack = stack
+        elif not stack:
+            # Previous search was completed and cache is still valid;
+            # no matching node.
+            return None
+
+        result = None
         while stack:
             node = stack.pop()
+            # add child elements to stack for continued searching
+            stack.extend([child for child in node.childNodes
+                          if child.nodeType in _nodeTypes_with_children])
             # check this node
-            if node.namespaceURI:
-                key = node.namespaceURI, node.localName
-            else:
-                key = node.tagName
-            info = self._elem_info.get(key)
+            info = self._get_elem_info(node)
             if info:
+                # We have to process all ID attributes before
+                # returning in order to get all the attributes set to
+                # be IDs using Element.setIdAttribute*().
                 for attr in node.attributes.values():
                     if attr.namespaceURI:
                         if info.isIdNS(attr.namespaceURI, attr.localName):
+                            self._id_cache[attr.value] = node
                             if attr.value == id:
-                                return node
-                            else:
+                                result = node
+                            elif not node._magic_id_nodes:
                                 break
                     elif info.isId(attr.name):
+                        self._id_cache[attr.value] = node
                         if attr.value == id:
-                            return node
-                        else:
+                            result = node
+                        elif not node._magic_id_nodes:
                             break
-            # didn't match; add child elements to stack
-            for child in node.childNodes:
-                if child.nodeType == Node.ELEMENT_NODE:
-                    stack.append(child)
-        return None
+                    elif attr._is_id:
+                        self._id_cache[attr.value] = node
+                        if attr.value == id:
+                            result = node
+                        elif node._magic_id_nodes == 1:
+                            break
+            elif node._magic_id_nodes:
+                for attr in node.attributes.values():
+                    if attr._is_id:
+                        self._id_cache[attr.value] = node
+                        if attr.value == id:
+                            result = node
+            if result is not None:
+                break
+        return result
 
     def getElementsByTagName(self, name):
         return _get_elements_by_tagName_helper(self, name, NodeList())
@@ -1459,6 +1776,7 @@ class Document(Node, DocumentLS):
         if n.nodeType == Node.ATTRIBUTE_NODE:
             element = n.ownerElement
             if element is not None:
+                is_id = n._is_id
                 element.removeAttributeNode(n)
         else:
             element = None
@@ -1475,6 +1793,8 @@ class Document(Node, DocumentLS):
             d['name'] = name
             if element is not None:
                 element.setAttributeNode(n)
+                if is_id:
+                    element.setIdAttributeNode(n)
         # It's not clear from a semantic perspective whether we should
         # call the user data handlers for the NODE_RENAMED event since
         # we're re-using the existing node.  The draft spec has been
@@ -1500,7 +1820,7 @@ def _clone_node(node, deep, newOwnerDocument):
                                                  node.nodeName)
         for attr in node.attributes.values():
             clone.setAttributeNS(attr.namespaceURI, attr.nodeName, attr.value)
-            a = clone.getAttributeNodeNS(attr.namespaceURI, attr.nodeName)
+            a = clone.getAttributeNodeNS(attr.namespaceURI, attr.localName)
             a.specified = attr.specified
 
         if deep:
@@ -1605,7 +1925,14 @@ def parseString(string, parser=None):
         return expatbuilder.parseString(string)
     else:
         from xml.dom import pulldom
-        return _do_pulldom_parse(pulldom.parseString, (string,), {'parser': parser})
+        return _do_pulldom_parse(pulldom.parseString, (string,),
+                                 {'parser': parser})
 
-def getDOMImplementation():
+def getDOMImplementation(features=None):
+    if features:
+        if isinstance(features, StringTypes):
+            features = domreg._parse_feature_string(features)
+        for f, v in features:
+            if not Document.implementation.hasFeature(f, v):
+                return None
     return Document.implementation
