@@ -2,7 +2,7 @@
 A library of useful helper classes to the saxlib classes, for the
 convenience of application and driver writers.
 
-$Id: saxutils.py,v 1.32 2002/08/13 09:28:51 afayolle Exp $
+$Id: saxutils.py,v 1.34 2002/10/28 18:15:43 fdrake Exp $
 """
 
 import os, urlparse, urllib2, types
@@ -15,6 +15,11 @@ try:
 except AttributeError: # 1.5 compatibility:UnicodeType not defined
     _StringTypes = [types.StringType]
 
+def __dict_replace(s, d):
+    """Replace substrings of a string using a dictionary."""
+    for key, value in d.items():
+        s = s.replace(key, value)
+    return s
 
 def escape(data, entities={}):
     """Escape &, <, and > in a string of data.
@@ -26,10 +31,24 @@ def escape(data, entities={}):
     data = data.replace("&", "&amp;")
     data = data.replace("<", "&lt;")
     data = data.replace(">", "&gt;")
-    for chars, entity in entities.items():
-        data = data.replace(chars, entity)
+    if entities:
+        data = __dict_replace(data, entities)
     return data
 
+def unescape(data, entities={}):
+    """Unescape &amp;, &lt;, and &gt; in a string of data.
+
+    You can unescape other strings of data by passing a dictionary as
+    the optional entities parameter.  The keys and values must all be
+    strings; each key will be replaced with its corresponding value.
+    """
+    data = data.replace("&lt;", "<")
+    data = data.replace("&gt;", ">")
+    if entities:
+        data = __dict_replace(data, entities)
+    # must do ampersand last
+    return data.replace("&amp;", "&")
+
 def quoteattr(data, entities={}):
     """Escape and quote an attribute value.
 
@@ -149,14 +168,45 @@ class ErrorRaiser:
 from xmlreader import AttributesImpl
 
 # --- XMLGenerator is the SAX2 ContentHandler for writing back XML
-try:
-    import codecs
-    def _outputwrapper(stream,encoding):
-        writerclass = codecs.lookup(encoding)[3]
-        return writerclass(stream)
-except ImportError: # 1.5 compatibility: fall back to do-nothing
-    def _outputwrapper(stream,encoding):
-        return stream
+import codecs
+
+def _outputwrapper(stream,encoding):
+    writerclass = codecs.lookup(encoding)[3]
+    return writerclass(stream)
+
+if hasattr(codecs, "register_error"):
+    def writetext(stream, text, entities={}):
+        stream.errors = "xmlcharrefreplace"
+        stream.write(escape(text, entities))
+        stream.errors = "strict"
+else:
+    def writetext(stream, text, entities={}):
+        text = escape(text, entities)
+        try:
+            stream.write(text)
+        except UnicodeError:
+            for c in text:
+                try:
+                    stream.write(c)
+                except UnicodeError:
+                    stream.write(u"&#%d;" % ord(c))
+
+def writeattr(stream, text):
+    countdouble = text.count('"')
+    if countdouble:
+        countsingle = text.count("'")
+        if countdouble <= countsingle:
+            entities = {'"': "&quot;"}
+            quote = '"'
+        else:
+            entities = {"'": "&apos;"}
+            quote = "'"
+    else:
+        entities = {}
+        quote = '"'
+    stream.write(quote)
+    writetext(stream, text, entities)
+    stream.write(quote)
 
 
 class XMLGenerator(handler.ContentHandler):
@@ -193,7 +243,8 @@ class XMLGenerator(handler.ContentHandler):
     def startElement(self, name, attrs):
         self._out.write('<' + name)
         for (name, value) in attrs.items():
-            self._out.write(' %s=%s' % (name, quoteattr(value)))
+            self._out.write(' %s=' % name)
+            writeattr(self._out, value)
         self._out.write('>')
 
     def endElement(self, name):
@@ -230,7 +281,8 @@ class XMLGenerator(handler.ContentHandler):
                 self._current_context[name[0]] = prefix
             else:
                 name = self._current_context[name[0]] + ":" + name[1]
-            self._out.write(' %s=%s' % (name, quoteattr(value)))
+            self._out.write(' %s=' % name)
+            writeattr(self._out, value)
         self._out.write('>')
 
     def endElementNS(self, name, qname):
@@ -246,7 +298,7 @@ class XMLGenerator(handler.ContentHandler):
         self._out.write('</%s>' % qname)
 
     def characters(self, content):
-        self._out.write(escape(content))
+        writetext(self._out, content)
 
     def ignorableWhitespace(self, content):
         self._out.write(content)
